/*
	File                 : DependenciesTest.cpp
	Project              : LabPlot
	Description          : Tests for dependencies
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2023 Stefan Gerlach <stefan.gerlach@uni.kn>

	SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "DependenciesTest.h"

QTEST_MAIN(DependenciesTest)
