/*
	File                 : GSLTest.cpp
	Project              : LabPlot
	Description          : Tests for Gsl
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2023 Martin Marmsoler <martin.marmsoler@gmail.com>

	SPDX-License-Identifier: GPL-2.0-or-later
*/
#include "GSLTest.h"

QTEST_MAIN(GslTest)
