/*
	File                 : GSLTest.h
	Project              : LabPlot
	Description          : Tests for GSL
	--------------------------------------------------------------------
	SPDX-FileCopyrightText: 2023 Martin Marmsoler <martin.marmsoler@gmail.com>

	SPDX-License-Identifier: GPL-2.0-or-later
*/

#ifndef GSLTEST_H
#define GSLTEST_H

#include "../../CommonTest.h"

class GslTest : public CommonTest {
	Q_OBJECT

private Q_SLOTS:
};

#endif // GSLTEST_H
